local CowTip = CowTip
local CowTip_LuaText = CowTip:NewModule("LuaText", "LibRockEvent-1.0", "LibRockHook-1.0", "LibRockTimer-1.0")
local self = CowTip_LuaText
--[[
-- Text-provider-code
--]]
local funcs = {}

local ScriptEnv
function CowTip_LuaText:SetScriptEnv(arg)
	ScriptEnv = arg
end

local mouseoverUnit = CowTip:GetMouseoverUnit()

function CowTip_LuaText:NewText(code)
	mouseoverUnit = CowTip:GetMouseoverUnit()
	local func
	local lua_string = 'return function(unit) if not unit then return \"\" end; '..code..' end'
	local create_func, err = loadstring(lua_string)
	if create_func then
		func = create_func()
		setfenv(func, ScriptEnv)
	else
		geterrorhandler()(err)
	end
	if not func then func = function() return "" end end
	funcs[code] = func
end

function CowTip_LuaText:RemoveText(code)
--	texts[code] = nil
	funcs[code] = nil
end

function CowTip_LuaText:GetText(code)
--	return texts[code]
	return string.format(funcs[code](CowTip:GetMouseoverUnit()))
end

--[[

--]]


local L = CowTip:L("CowTip_Text")

local leftGameTooltipStrings = CowTip.leftGameTooltipStrings
local rightGameTooltipStrings = CowTip.rightGameTooltipStrings

local _G = _G
local IsShiftKeyDown = _G.IsShiftKeyDown
local ChatFontNormal = _G.ChatFontNormal
local CreateFrame = _G.CreateFrame
local UIParent = _G.UIParent
local GameTooltip = _G.GameTooltip
local ipairs = _G.ipairs
local GetNumFactions = _G.GetNumFactions
local GetFactionInfo = _G.GetFactionInfo
local UnitPlayerControlled = _G.UnitPlayerControlled
local UnitIsPlayer = _G.UnitIsPlayer
local UnitFactionGroup = _G.UnitFactionGroup
local UnitIsConnected = _G.UnitIsConnected
local UnitIsVisible = _G.UnitIsVisible
local UnitIsPVP = _G.UnitIsPVP
local UnitExists = _G.UnitExists

local kwargs = { unit = 'mouseover' }

function CowTip_LuaText:OnInitialize()
	self.db = CowTip:GetDatabaseNamespace("LuaText")
	CowTip:SetDatabaseNamespaceDefaults("LuaText", "profile", {
		lefts = {
			['*'] = "",
			[[local name = UnitPVPName(unit) or Name(unit)
local _, class = UnitClass(unit)
local color = "%1"
if UnitIsPlayer(unit) then
  color = Colorize("%1", ClassColor(unit))
else
  color = Colorize("%1", HostileColor(unit))
end
name = name:gsub(Name(unit), color)
return name
]],
			[[local guild, rankname, rankindex = GetGuildInfo(unit)
if not guild then return "" end
local pguild, _, prankindex = GetGuildInfo("player")
if pguild == guild then
  guild = ("|cff00cc00<%s>|r"):format(guild)
  local rankcolor
  if rankindex > prankindex then
    rankcolor = "|cffff0000%s|r"
  elseif rankindex < prankindex then
    rankcolor = "|cff00ff00%s|r"
  else
    rankcolor = "|cffffff00%s|r"
  end
  rankname = rankcolor:format(rankname)
else
  guild = ("<%s>"):format(guild)
end
return "%s (%d) of %s", rankname, rankindex, guild
]],
			("return \"%s\""):format(TARGET),
[[local level, race = Level(unit), SmartRace(unit)
local class = ""
if UnitIsPlayer(unit) then
  class = Colorize(Class(unit), ClassColor(unit))
else
  class = ""
end
level = Colorize(level, DifficultyColor(unit))
local diff = ShortClassification(unit)
return "Level %s%s %s %s", level, diff, race, class
]]
--			"[if [IsPlayer or [IsEnemy and not IsPet]] then ClassColor][NameRealm]", -- [1]
--[[			L["Target:"], -- [2]
			L["Guild:"], -- [3]
			L["Guild rank:"], -- [4]
			L["Level:"], -- [5]
			L["Race:"], -- [6]
			L["Class:"], -- [7]
			L["Faction:"], -- [8]
			L["Status:"], -- [9]
			L["Health:"], -- [10]
			"[if HasMP then TypePower ':']", -- [11]
			L["Experience:"], -- [12]
			L["Threat:"], -- [13]
			L["Happiness:"], -- [14]
			L["Zone:"], -- [15]
			(_G.TALENTS or "Talents") .. ":", -- [16]--]]
		},
		rights = {
			['*'] = "",
			"",
			"",
			[[return UnitName(unit .. "target") or NONE]],
	--[[		"", -- [1]
			"[if Target:IsPlayer or (Target:IsEnemy and not Target:IsPet) then ClassColor(unit=Target)][if IsUnit('player', Target) then '<<" .. L["You"] .. ">>' else Target:NameRealm or '" .. L["None"] .. "']", -- [2]
			"[if Guild and Guild = 'player':Guild then Green][Guild:Angle]", -- [3]
			"[GuildRank]", -- [4]
			"[Classification] [Level]", -- [5]
			"[if IsPlayer then Race else Creature]", -- [6]
			"[if IsPlayer or [IsEnemy and not IsPet] then Class:ClassColor]", -- [7]
			"[Faction]", -- [8]
			"[Status]", -- [9]
			"[FractionalHP:Short] [PercentHP:Percent:Paren]", -- [10]
			"[if HasMP then FractionalMP:Short ' ' PercentMP:Percent:Paren]", -- [11]
			"[if XP ~= 0 then FractionalXP:Short ' ' PercentXP:Percent:Paren]", -- [12]
			"[if HasThreat then FractionalThreat:Short ' ' PercentThreat:Percent:Paren]", -- [13]
			"[if IsUnit('pet') then HappyText]", -- [14]
			"[Zone]", -- [15]
			"[if TalentSpec then TalentTree ' ' TalentSpec:Paren]", -- [16]--]]
		}
	})
end

local opt
local function makeOptions()
	if not opt then
		opt = {
			name = L["Lua Line texts"],
			desc = L["Here you can describe the text on the assorted lines of the tooltip in Lua."],
			type = 'group',
			args = {},
		}
	
		local function getLeft(i)
			return self.db.profile.lefts[i] or ""
		end
		local function getRight(i)
			return self.db.profile.rights[i] or ""
		end
		local function setText(side, i, value)
			local old = self.db.profile[side][i]
			self:NewText(value)
			self.db.profile[side][i] = value
			if old then
				self:RemoveText(old)
			end
		end
		
		local function setLeft(i, value)
			setText("lefts", i, value)
		end
		local function setRight(i, value)
			setText("rights", i, value)
		end
		
		for i = 1,30 do
			local line_opt = {
				name = L["Line %d"]:format(i),
				desc = "",
				type = 'group',
				groupType = 'inline',
				args = {
					left = {
						name = L["Left"],
						desc = L["Left"],
						type = 'string',
						multiline = true,
						get = getLeft,
						set = setLeft,
						passValue = i,
						usage = "<code>",
					},
					right = {
						name = L["Right"],
						desc = L["Right"],
						type = 'string',
						multiline = true,
						get = getRight,
						set = setRight,
						passValue = i,
						usage = "<code>",
					},
				},
				order = i,
			}
			opt.args["line" .. i] = line_opt
		end -- for i = 1,9
		self:RegisterCowTipOption(opt)
	end -- if not opt
end

function CowTip_LuaText:OnEnable()
	CowTip:GetModule("Text"):ToggleActive(false)
	for k,v in pairs(self.db.profile.lefts) do
		self:NewText(v, {})
	end
	for k,v in pairs(self.db.profile.rights) do
		self:NewText(v, {})
	end
	
	makeOptions()
	
	self:AddEventListener("UPDATE_FACTION")
	
	self:AddHook(GameTooltip, "GetUnit", "GameTooltip_GetUnit")
	self:UPDATE_FACTION()
end

function CowTip_LuaText:OnDisable()
	wipe(funcs)
end

local factionList = {}

function CowTip_LuaText:UPDATE_FACTION()
	for i = 1, GetNumFactions() do
		local name = GetFactionInfo(i)
		factionList[name] = true
	end
end

--[[
How normal tooltips look:
{Name}
{[Guild] : ~~~}
Level {Num} {[(Classification)]}
{NPC? [Faction] : ~~~}
{Offline? Offline : !IsVisible? [Zone] : IsPVP? [PvP] : ~~~}
]]

local lastNumRegistered = 0
local linesToAdd = {}
local linesToAddR = {}
local linesToAddG = {}
local linesToAddB = {}
local linesToAddRight = {}
local linesToAddRightR = {}
local linesToAddRightG = {}
local linesToAddRightB = {}

local lastUnitName, lastUnit

function CowTip_LuaText:RefixLines(lastLine)
	for i = 1, lastNumRegistered do
--		DogTag:RemoveFontString(leftGameTooltipStrings[i])
--		DogTag:RemoveFontString(rightGameTooltipStrings[i])
	end
	lastNumRegistered = 0
	local LinesAdded = 0

	for i = lastLine, GameTooltip:NumLines() do
		local left = leftGameTooltipStrings[i]
		local j = i - lastLine + 1
		linesToAdd[j] = left:GetText()
		local r, g, b = left:GetTextColor()
		linesToAddR[j] = r
		linesToAddG[j] = g
		linesToAddB[j] = b
		local right = rightGameTooltipStrings[i]
		if right:IsShown() then
			linesToAddRight[j] = right:GetText()
			local r, g, b = right:GetTextColor()
			linesToAddRightR[j] = r
			linesToAddRightG[j] = g
			linesToAddRightB[j] = b
		end
		LinesAdded = LinesAdded + 1
	end
	
	GameTooltip:ClearLines()
	local db = self.db.profile
	local lefts, rights = db.lefts, db.rights
	local num = 0
	kwargs.unit = CowTip:GetMouseoverUnit()
	for i = 1, 30 do
		local left, right = lefts[i], rights[i]
		local left_ev, right_ev
		if left and left:trim() == "" then
			left = nil
		else
			--left_ev = tostring(DogTag:Evaluate(left, "Unit", kwargs) or '')
			left_ev = tostring(self:GetText(left))
			if left_ev:trim() == "" then
				left_ev = nil
			end
		end
		if right and right:trim() == "" then
			right = nil
		else
--			right_ev = tostring(DogTag:Evaluate(right, "Unit", kwargs) or '')
			right_ev = tostring(self:GetText(right))
			if right_ev:trim() == "" then
				right_ev = nil
			end
		end
		if right then
			if right_ev or (left and left_ev and left:find("%[")) then
				num = num + 1
				GameTooltip:AddDoubleLine(left_ev or ' ', right_ev, 1, 1, 1, 1, 1, 1)
				local leftFS = leftGameTooltipStrings[num]
				local rightFS = rightGameTooltipStrings[num]
--				DogTag:AddFontString(leftFS, GameTooltip, left or ' ', "Unit", kwargs)
--				DogTag:AddFontString(rightFS, GameTooltip, right, "Unit", kwargs)
			end
		elseif left then
			if left_ev then
				num = num + 1
				GameTooltip:AddLine(left_ev, 1, 1, 1, true)
				local leftFS = leftGameTooltipStrings[num]
				leftFS:SetTextColor(1, 1, 1)
--				DogTag:AddFontString(leftFS, GameTooltip, left, "Unit", kwargs)
			end
		end
	end
	if lastNumRegistered < num then
		lastNumRegistered = num
	end

	for i = 1, LinesAdded do
		local left = linesToAdd[i]
		local right = linesToAddRight[i]
		if right then
			GameTooltip:AddDoubleLine(left, right, linesToAddR[i], linesToAddG[i], linesToAddB[i], linesToAddRightR[i], linesToAddRightG[i], linesToAddRightB[i])
		else
			GameTooltip:AddLine(left, linesToAddR[i], linesToAddG[i], linesToAddB[i], true)
		end
		linesToAdd[i] = nil
		linesToAddR[i] = nil
		linesToAddG[i] = nil
		linesToAddB[i] = nil
		linesToAddRight[i] = nil
		linesToAddRightR[i] = nil
		linesToAddRightG[i] = nil
		linesToAddRightB[i] = nil
	end
end

function CowTip_LuaText:GameTooltip_GetUnit(this)
	if not lastUnitName then
		return self.hooks[GameTooltip].GetUnit(GameTooltip)
	end
	return lastUnitName, lastUnit
end

local LEVEL_start = "^" .. _G.LEVEL
local PVP = _G.PVP

function CowTip_LuaText:OnTooltipSetUnit()
	lastUnitName, lastUnit = self.hooks[GameTooltip].GetUnit(GameTooltip)
	local num = 2
	local text_2 = leftGameTooltipStrings[2]:GetText()
	if not text_2 then
		num = num - 1
	elseif not text_2:find(LEVEL_start) then
		-- has a guild
		num = num + 1
	end
	local mouseover_unit = CowTip:GetMouseoverUnit()
	if not UnitPlayerControlled(mouseover_unit) and not UnitIsPlayer(mouseover_unit) then
		-- npc, faction line
		local factionText = leftGameTooltipStrings[num+1]:GetText()
		if factionText == PVP then
			factionText = nil
		end
		if factionText and (factionList[factionText] or UnitFactionGroup(mouseover_unit)) then
			num = num + 1
		end
	end
	if not UnitIsConnected(mouseover_unit) or not UnitIsVisible(mouseover_unit) or UnitIsPVP(mouseover_unit) then
		-- offline, zone, or pvp
		num = num + 1
	end
	
	self:RefixLines(num+1)
	self:AddRepeatingTimer("CowTip_LuaText_UpdateTexts", 1, "UpdateTexts")
end

function CowTip_LuaText:UpdateTexts()
	if not lastUnit or not UnitExists(mouseover_unit) then
		return
	end
	self:RefixLines(lastNumRegistered+1)
	
	GameTooltip:Show()
end

function CowTip_LuaText:OnTooltipHide()
	for i = 1, lastNumRegistered do
--		DogTag:RemoveFontString(leftGameTooltipStrings[i])
--		DogTag:RemoveFontString(rightGameTooltipStrings[i])
	end
	lastNumRegistered = 0
	lastUnitName, lastUnit = nil, nil
	
	self:RemoveTimer("CowTip_LuaText_UpdateTexts")
end

local function CleanCode(text)
--	text = DogTag:CleanCode(text)
	text = text:gsub("%s%s*", " ")
	return text
end

local function fixColor()
	GameTooltipTextLeft1:SetTextColor(1, 1, 1)
end

function CowTip_LuaText:PostTooltipSetUnit()
	self:AddTimer(0, fixColor)
end
